#include <string.h>
/*
  My attempt at putting the Machine/Section control code all into a class
    by Matt Elias 240219

  Supports control of regular Arduino pins as well as PCA9555 I2C port expanders as used on the AiO v5.0a Proto
    - two init() functions, 1st for Arduino pins and the 2nd for PCA9555/v5.0a control
    - currently only supports Teensy NativeEthernet UDP


  To to:
    - add code so that watchdog works
*/


#include "EEPROM.h"
#include "wiring.h"
#include "core_pins.h"
#include <stdint.h>
#include "Arduino.h"
#include <NativeEthernet.h>
#include <NativeEthernetUdp.h>
#include <IPAddress.h>
#include "clsPCA9555.h"

class MACHINE
{
private:
  int16_t eeAddr = -1;                            // -1 defaults to no EEPROM saving/loading
  bool eeLoadedAtStartup = 0;
  uint8_t numOutputPins = 0;                      // 0 defaults to no direct Arduino pin control
  const static uint8_t maxOutputPins = 24;        // 24 pins can be configured in AoG (Machine Pin Config PGN), 64 sections currently the max supported by AoG
  uint8_t outputPinNumbers[1 + maxOutputPins];    // all the '1 +' are to eliminate 0 indexing so that outputPinNumber[1] is pin 1 (the first pin)
  bool sectionStates[1 + 64];                     // store state of up to 64 sections, from AOG

  // read value from Machine data and set 1 or zero according to list, 21 different function types
  bool functionStates[1 + 21] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 };

  PCA9555* pcaOutputs = NULL;
  uint8_t pcaOutputPinNumbers[1 + 8];

  elapsedMillis watchdogTimer;
  const uint8_t LOOP_TIME = 200;                //5hz
  const uint16_t watchdogTimeoutPeriod = 4000;         // ms, originally was 20 update cycles (4 secs)
  //uint8_t watchdogCount = watchdogTimeout;      //make sure we are talking to AOG

  const String functionNames[1 + 21] = { "",
      "S01", "S02", "S03", "S04", "S05", "S06", "S07", "S08",
      "S09", "S10", "S11", "S12", "S13", "S14", "S15", "S16",
      "HydUp", "HydDn", "TramR", "TramL", "GeoStop"
  };

public:
  #define EE_IDENT 11  // change to force EE to reset to defaults

  struct Config {
    uint8_t raiseTime = 2;
    uint8_t lowerTime = 4;
    uint8_t hydLiftEnable = false;
    uint8_t isPinActiveHigh = 0;   // if zero, active low (default)
  
    uint8_t user1 = 0; //user defined values set in machine tab
    uint8_t user2 = 0;
    uint8_t user3 = 0;
    uint8_t user4 = 0;

    uint8_t pinFunction[1 + 24] = { 0,1,2,3,4,5,6,7,8,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 };
  };  Config config;   //4 bytes
  /* pin function numbers as below, assigned to pinFunction[]
  1-16   Section 1-16
  17,18  Hyd Up, Hyd Down,
  19,20  Tramline Right, Left
  21     Geo Stop  */
  // 24 is max pins currently supported in AoG


  //The variables used for storage
  //uint8_t relayHi = 0, relayLo = 0
  uint8_t tramline = 0, uTurn = 0, hydLift = 0, geoStop = 0;
  uint8_t raiseTimer = 0, lowerTimer = 0, lastTrigger = 0;
  bool isRaise = false, isLower = false;
  float gpsSpeed;
  uint8_t debugLevel = 5;   // 0 - debug prints OFF
                            // 1 - alerts/errors only
                            // 2 - init info
                            // 3 - config PGN announcements & data (only sent when changing settings in AOG)
                            // 4 - all PGN announcements
                            // 5 - all PGN data
  uint8_t debugMask = B0000;


  MACHINE(void) {}
  ~MACHINE(void) {}

  void init(uint8_t* _outputPinNumbers, uint8_t _numOutputPins = 0, int16_t _eeAddr = -1)
  {
    numOutputPins = min(_numOutputPins, maxOutputPins);   // limit to maxOutputPins (24 or 64)
    //Serial.print("\r\nnumOutputPins:"); Serial.print(numOutputPins);
    eeAddr = _eeAddr;
    loadFromEeprom();

    if (numOutputPins > 0) {
      for (uint8_t i = 1; i <= numOutputPins; i++) {
        outputPinNumbers[i] = _outputPinNumbers[i - 1];
        pinMode(outputPinNumbers[i], OUTPUT);
        digitalWrite(outputPinNumbers[i], !config.isPinActiveHigh);
      }
    }
  }

  void init(PCA9555* _pcaOutputs, uint8_t* _outputPins, int16_t _eeAddr = -1)
  {
    pcaOutputs = _pcaOutputs;
    eeAddr = _eeAddr;
    loadFromEeprom();

    for (uint8_t i = 1; i <= 8; i++){
      pcaOutputPinNumbers[i] = _outputPins[i - 1];
      pcaOutputs->pinMode(pcaOutputPinNumbers[i], OUTPUT);
      pcaOutputs->digitalWrite(outputPinNumbers[i], config.isPinActiveHigh);
    }
  }

  void loadFromEeprom()
  {
    if (eeLoadedAtStartup) return;
    if (eeAddr < 0) return;

    uint8_t EEread;
    EEPROM.get(eeAddr + 0, EEread);        // read version identifier
    if (EEread != EE_IDENT) {              // check on first start and write EEPROM
      EEPROM.put(eeAddr + 0, EE_IDENT);
      EEPROM.put(eeAddr + 2, config);      // +2 to leave room for EE_IDENT
      Serial.print("\r\n\n* Machine config reset to default (new EEPROM version) *");
    } else {
      EEPROM.get(eeAddr + 2, config);
      Serial.print("\r\n\nMachine config loaded from EEPROM");
    }
    printConfig();
    printPinConfig();
    eeLoadedAtStartup = true;
  }

  void saveToEeprom()
  {
    if (eeAddr < 0) return;
    EEPROM.put(eeAddr + 2, config);
    if (debugLevel > 1) Serial.print("\r\nNew Machine config saved to EEPROM");
  }

  void printConfig()
  {
    if (debugLevel > 1) {
      Serial << "\r\n- raiseTime: " << config.raiseTime;
      Serial << "\r\n- lowerTime: " << config.lowerTime;
      Serial << "\r\n- relayActiveHigh: " << config.isPinActiveHigh;
      Serial << "\r\n- hydLiftEnable: " << config.hydLiftEnable;
      Serial << "\r\n- user1: " << config.user1;
      Serial << "\r\n- user2: " << config.user2;
      Serial << "\r\n- user3: " << config.user3;
      Serial << "\r\n- user4: " << config.user4;
    }
  }

  void printPinConfig()
  {
    if (debugLevel > 1) {
      for (uint8_t i = 1; i < uint8_t(sizeof(config.pinFunction)); i++) {
        Serial << "\r\n- Pin " << (i < 10 ? " " : "") << i << ": " << (config.pinFunction[i] < 10 ? " " : "") << config.pinFunction[i] << " " << functionNames[config.pinFunction[i]];
      }
    }
  }


  void updatePinStates() {
    // set pins according to functionStates unless watchdog has timed out, then set to !isPinActiveHigh

    if (numOutputPins > 0) {
      //if (debugLevel > 3) Serial.print("\r\nPin outputs ");
      for (uint8_t i = 1; i <= numOutputPins; i++) {
        digitalWrite(outputPinNumbers[i], functionStates[config.pinFunction[i]] == config.isPinActiveHigh);                     // ==, XOR
        //if (debugLevel > 3) Serial << i << ":" << (functionStates[config.pinFunction[i]] == config.isPinActiveHigh) << " ";
      }
    }

    if (pcaOutputs != NULL) {
      //if (debugLevel > 3) Serial.print("\r\nPCA outputs ");
      for (uint8_t i = 1; i <= 8; i++) {       // AiO v5.0a has 8 PCA9555 outputs
        if (config.pinFunction[i] > 0) {
          pcaOutputs->digitalWrite(pcaOutputPinNumbers[i], !(functionStates[config.pinFunction[i]] == config.isPinActiveHigh));   // NXOR
          //if (debugLevel > 3) Serial << i << ":" << !(functionStates[config.pinFunction[i]] == config.isPinActiveHigh) << " ";
        }
      }
    }
    /*Serial.println();
    for (byte i = 0; i < 24; i++){
      Serial.printf("%2i ", config.pinFunction[i]);
    }
    Serial.println();
    for (byte i = 0; i < 24; i++){
      Serial.printf("%2i ", (config.pinFunction[i] > 0 ? functionStates[config.pinFunction[i]] : -1));
    }
    Serial.println();
    for (byte i = 0; i < 22; i++){
      Serial.printf("%2i ", functionStates[i]);
    }*/
    //Serial << "\r\noutput update delay: " << updateDelayTimer;
    //if (config.pinFunction[0]) digitalWrite(4, functionStates[config.pinFunction[0]]);
  }


  void watchdogCheck()
  {
    if (watchdogTimer > watchdogTimeoutPeriod)    // watchdogTimer reset with Machine Data PGN, should be 64 Section instead or both?
    {    
      updatePinStates();
      watchdogTimer = 0;            // only output timed out OFF every watchdogTimeoutPeriod
    }
  }

  // output update triggered by PGN from AOG for quicker section response, watchdogCheck() looks for comms timeout
  void updateStates() // uint8_t* _data, uint8_t _size) //EthernetUDP* _udp, IPAddress _dIP, uint16_t _dPort)
  {
    //Serial << "\r\n" << millis() << " machine update " << _dIP << ":" << _dPort;

    //hydraulic lift
    if (config.hydLiftEnable) {
      if (hydLift != lastTrigger && (hydLift == 1 || hydLift == 2))
      {
        lastTrigger = hydLift;
        lowerTimer = 0;
        raiseTimer = 0;

        switch (hydLift) {
          case 1:   //lower
            lowerTimer = config.lowerTime * 5;      // secs * 5 update cycles per second (5hz)
            break;
          case 2:   //raise
            raiseTimer = config.raiseTime * 5;
            break;
        }
      }

      //countdown if not zero, UP only (?)
      if (raiseTimer) {
        raiseTimer--;
        lowerTimer = 0;
      }
      if (lowerTimer) lowerTimer--;

      //if anything wrong, shut off hydraulics, reset last
      if (hydLift != 1 && hydLift != 2) { //|| gpsSpeed < 2)
        lowerTimer = 0;
        raiseTimer = 0;
        lastTrigger = 0;
      }

      isLower = isRaise = false;          // set both OFF
      if (lowerTimer) isLower = true;     // if lower is active, set ON
      if (raiseTimer) isRaise = true;     // if raise is active, set ON
    }
    else {      // hyd lift is disabled, make sure any hyd lift pins are OFF
      isLower = false;
      isRaise = false;

    }

    /*float speedPulse = gpsSpeed * 36.1111;
    if (speedPulse > 2.0) {                    // less then 2 hz (0.055 km/hr) doesn't work well with tone()
      tone(speedPulsePin, uint16_t(speedPulse));
    } else {
      noTone(speedPulsePin);
    }*/

    //Load the current output states for section 1-16 only from 64 Sections PGN
    for (uint8_t i = 1; i <= 16; i++) {
        functionStates[i] = sectionStates[i];    // copy from 64 Sections PGN
    }

    // Hydraulics
    functionStates[17] = isLower;
    functionStates[18] = isRaise;

    //Tram
    functionStates[19] = bitRead(tramline, 0); //right
    functionStates[20] = bitRead(tramline, 1); //left

    //GeoStop
    functionStates[21] = geoStop;

    updatePinStates();

    // *** Sending PGN_237 isn't necessary/doesn't do anything?

    // generic "from machine" PGN template
    //uint8_t PGN_237[14] = { 0x80, 0x81, 0x7B, 237, 8, 1, 2, 3, 4, 0, 0, 0, 0, 0xCC };

    //checksum
    /*int16_t CK_A = 0;
    for (uint8_t i = 2; i < sizeof(PGN_237) - 1; i++) {
      CK_A = (CK_A + PGN_237[i]);
    }
    PGN_237[sizeof(PGN_237) - 1] = CK_A;*/

    //off to AOG
    //sendUdp(PGN_237, sizeof(PGN_237), _dIP, _dPort, _udp);
    //_udp->beginPacket(_dIP, _dPort);
    //_udp->write(PGN_237, sizeof(PGN_237));
    //_udp->endPacket();
  }


  bool parsePGN(uint8_t *pgnData, uint8_t len)
  {
    if (len < 5) return false;
    // should check CRC here too???
    if (pgnData[0] != 0x80 || pgnData[1] != 0x81 || pgnData[2] != 0x7F) return false;    // skip the rest if the first three bytes are NOT AoG headers


    if (pgnData[3] == 229 && len == 16)                    // 0xE5 (229) - 64 Section Data
    {                                                      // use this instead of relayLo/Hi from other PGNs because it works for zones/groups too
      
      //if (debugLevel > 3) { Serial.print("\r\n0x"); Serial.print(pgnData[3], HEX); Serial.print((String)" (" + pgnData[3] + ") - "); }
      //if (debugLevel > 3) Serial.print("64 Section Data");

      //if (debugLevel > 3) Serial.println();
      for (uint8_t j = 0; j < 8; j++) {
        //if (debugLevel > 3) Serial << j + 1 << ":";
        for (uint8_t i = 0; i < 8; i++) {
          sectionStates[1 + i + j * 8] = bitRead(pgnData[5 + j], i);
          //if (debugLevel > 3) Serial.print(sectionStates[i + j * 8]);
        }
        //if (debugLevel > 3) Serial.print(" ");
      }

      //if (debugLevel > 3) Serial.println(); 
      updateStates();
      return true;
    }


    else if (pgnData[3] == 235 && len == 39)               // 0xEB (235) - Section Dimensions
    {
      if (debugLevel > 2) { Serial.print("\r\n0x"); Serial.print(pgnData[3], HEX); Serial.print((String)" (" + pgnData[3] + ") - "); }
      if (debugLevel > 2) Serial.print("Section Dimensions");
      
      // parse section dims here
      
      if (debugLevel > 2) Serial.println(); 
      return true;
    }


    else if (pgnData[3] == 236 && len == 30)               // 0xEC (236) - Machine Pin Config
    {
      if (debugLevel > 2) { Serial.print("\r\n0x"); Serial.print(pgnData[3], HEX); Serial.print((String)" (" + pgnData[3] + ") - "); }
      if (debugLevel > 2) Serial.print("Machine Pin Config");
      for (uint8_t i = 5; i < min(len - 1, uint8_t(sizeof(config.pinFunction) + 5)); i++) {
        config.pinFunction[i - 4] = pgnData[i];
      }
      if (debugLevel > 2) printPinConfig();
      saveToEeprom();

      if (debugLevel > 2) Serial.println();
      return true;
    }


    else if (pgnData[3] == 238 && len == 14)                // 0xEE (238) - Machine Config
    {
      if (debugLevel > 2) { Serial.print("\r\n0x"); Serial.print(pgnData[3], HEX); Serial.print((String)" (" + pgnData[3] + ") - "); }
      if (debugLevel > 2) Serial.print("Machine Config");
      config.raiseTime = pgnData[5];
      config.lowerTime = pgnData[6];
      //config.hydLiftEnable = pgnData[7];    // not used?

      uint8_t set0 = pgnData[8];  // setting0
                                  // bit 0: relayActiveHigh
                                  // bit 1: hydLiftEnable
      config.isPinActiveHigh = bitRead(set0, 0) ? 1 : 0;
      config.hydLiftEnable   = bitRead(set0, 1) ? 1 : 0;

      config.user1 = pgnData[9];
      config.user2 = pgnData[10];
      config.user3 = pgnData[11];
      config.user4 = pgnData[12];

      //Serial << "\r\n- set0: " << set0 << " " << (bitRead(set0, 3) ? 1 : 0) << (bitRead(set0, 2) ? 1 : 0) << (bitRead(set0, 1) ? 1 : 0) << (bitRead(set0, 0) ? 1 : 0);
      if (debugLevel > 2) printConfig();
      saveToEeprom();
      //rebootFunc();    // from old code, is there any reason to reboot?

      if (debugLevel > 2) Serial.println();
      return true;
    }


    else if (pgnData[3] == 239 && len == 14)                // 0xEF (239) - Machine Data
    {
      if (debugLevel > 3) { Serial.print("\r\n0x"); Serial.print(pgnData[3], HEX); Serial.print((String)" (" + pgnData[3] + ") - "); }
      if (debugLevel > 3) Serial.print("Machine Data");
      
      uTurn = pgnData[5];

      gpsSpeed = (float)pgnData[6] * 0.1;     // gpsSpeed is 10x actual speed
      //speedPulseUpdate();

      hydLift = pgnData[7];
      tramline = pgnData[8];  // bit 0 is right bit 1 is left
      geoStop = (pgnData[9] > 0 ? 1 : 0);

      //relayLo = pgnData[11];  // use 64 Sections PGN data instead
      //relayHi = pgnData[12];

      if (debugLevel > 4) {
        Serial << "\r\nuTurn: " << uTurn;
        Serial << "\r\ngpsSpeed: " << gpsSpeed;
        Serial << "\r\nhydLift(up/down): " << hydLift;
        Serial << "\r\ntramline: " << pgnData[8] << " " << tramline;
        Serial << "\r\ngeoStop: " << pgnData[9] << " " << geoStop;
        Serial.print("\r\nsec 1-8: "); printBinary(pgnData[11]);// Serial.print(" "); Serial.print(relayLo,BIN);
        Serial.print(" sec 9-16: "); printBinary(pgnData[12]);// Serial.print(" "); Serial.print(relayHi,BIN);
      } else if (debugLevel > 3) {
        Serial.print("\r\n0:"); printBinary(pgnData[11]);
        Serial.print(" 1:"); printBinary(pgnData[12]);
      }

      watchdogTimer = 0;   //reset watchdog

      // updating outputs here would be slightly quicker response then waiting for old update loop to trigger, at times the delay is almost 200ms
      updateStates();

      if (debugLevel > 3) Serial.println();
      return true;
    }

    else
    {
      return false;   // no matching PGN, return false for further PGN processing
    }

    return false;     // should never reach this
  }

  void printBinary(uint8_t var) {
    for (uint8_t bit = 0; bit < 8; bit++){
      Serial.print(bitRead(var, bit));
    }
  }

  void calculateAndSetCRC(uint8_t myMessage[], uint8_t myLen) {
    if (myLen <= 2 ) return;

    uint8_t crc = 0;
    for (byte i = 2; i < myLen - 1; i++) {
      crc = (crc + myMessage[i]);
    }
    myMessage[myLen - 1] = crc;
  }


};
